use std::fmt::Display;
use std::io::Write;
use std::mem::size_of;
use std::{
    error::Error,
    fs::File,
    io::{Read, Seek, SeekFrom},
};

pub trait PointCache {
    type Frame: Frame;
    fn read_frames(
        infile: File,
    ) -> Result<
        Box<dyn Iterator<Item = Result<<Self as PointCache>::Frame, Box<dyn Error>>>>,
        Box<dyn Error>,
    >;
    fn read_animation_info(infile: File) -> Result<AnimationInfo, Box<dyn Error>>;
    fn map_to<F: FnMut(<Self as PointCache>::Frame) -> Result<MDDFrame, Box<dyn Error>>>(
        infile: File,
        outfile: File,
        op: F,
    ) -> Result<(), Box<dyn Error>>;
    fn read(infile: File) -> Result<Animation<Self>, Box<dyn Error>> {
        Ok(Animation {
            frames: Self::read_frames(infile.try_clone()?)?,
            info: Self::read_animation_info(infile)?,
        })
    }
}

pub trait Frame {
    fn point_idx(&self) -> usize;
    fn point(&self) -> Point;
    fn with_point(self, point: Point) -> Self;
}

pub struct MDDSeekableFile;

#[repr(C)]
#[derive(Debug, Default, Clone, Copy, PartialEq)]
pub struct Point {
    pub x: f32,
    pub y: f32,
    pub z: f32,
}

#[derive(Debug, PartialEq)]
pub struct AnimationInfo {
    pub total_frames: usize,
    pub total_points: usize,
    pub start_frame: usize,
    pub end_frame: usize,
    // MDD file format supports varying FPS , but Blender's exporter does not
    // FPS is only correct if it's not varying, see individual frame times for more granularity
    pub fps: f32,
}

#[derive(Debug)]
pub struct EmptyAnimationError;

impl Display for EmptyAnimationError {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "EmptyAnimationError")
    }
}
impl Error for EmptyAnimationError {}

pub struct Animation<PC: PointCache + ?Sized> {
    pub frames: Box<dyn Iterator<Item = Result<<PC>::Frame, Box<dyn Error>>>>,
    pub info: AnimationInfo,
}

#[derive(Debug, Clone, PartialEq)]
pub struct MDDFrame {
    pub frame_idx: usize,
    pub point_idx: usize,
    pub time: f32,
    pub point: Point,
}

impl Frame for MDDFrame {
    fn point_idx(&self) -> usize {
        self.point_idx
    }
    fn point(&self) -> Point {
        self.point
    }
    fn with_point(self, point: Point) -> Self {
        MDDFrame { point, ..self }
    }
}

impl MDDSeekableFile {
    pub fn read_header(infile: &mut File) -> Result<(i32, i32), Box<dyn Error>> {
        infile.seek(SeekFrom::Start(0))?;
        let mut buff = [0u8; size_of::<i32>()];
        infile.read_exact(&mut buff)?;
        let total_frames = i32::from_be_bytes(buff);
        infile.read_exact(&mut buff)?;
        let total_points = i32::from_be_bytes(buff);

        return Ok((total_frames, total_points));
    }
    fn write_header(
        outfile: &mut File,
        total_frames: i32,
        total_points: i32,
    ) -> Result<(), Box<dyn Error>> {
        outfile.seek(SeekFrom::Start(0))?;
        outfile.write_all(&total_frames.to_be_bytes())?;
        outfile.write_all(&total_points.to_be_bytes())?;
        Ok(())
    }
}

#[derive(Debug)]
struct ExaustedPointsUnexpectedly;
impl Display for ExaustedPointsUnexpectedly {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "ExaustedPointsUnexpectedly")
    }
}
impl Error for ExaustedPointsUnexpectedly {}

impl PointCache for MDDSeekableFile {
    type Frame = MDDFrame;
    fn read_frames(
        mut infile: File,
    ) -> Result<
        Box<dyn Iterator<Item = Result<<Self as PointCache>::Frame, Box<dyn Error>>>>,
        Box<dyn Error>,
    > {
        let (total_frames, total_points) = MDDSeekableFile::read_header(&mut infile)?;

        let header_size: u64 = (size_of::<i32>() * 2).try_into()?;
        let f32_size: u64 = size_of::<f32>().try_into()?;
        let mut buff = [0u8; size_of::<f32>()];
        let frames = Box::new(
            (0..((total_frames * total_points) as u64)).map::<Result<_, Box<dyn Error>>, _>(
                move |count| {
                    let frame_idx: usize = (count / (total_points as u64)).try_into()?;
                    let time_offset: u64 = f32_size * (frame_idx as u64);
                    let time_size: u64 = f32_size * (total_frames as u64);
                    let data_offset: u64 = f32_size * count * 3;

                    infile.seek(SeekFrom::Start(header_size + time_offset))?;
                    infile.read_exact(&mut buff)?;
                    let time = f32::from_be_bytes(buff);

                    infile.seek(SeekFrom::Start(header_size + time_size + data_offset))?;
                    infile.read_exact(&mut buff)?;
                    let x = f32::from_be_bytes(buff);
                    infile.read_exact(&mut buff)?;
                    let y = f32::from_be_bytes(buff);
                    infile.read_exact(&mut buff)?;
                    let z = f32::from_be_bytes(buff);
                    let point = Point { x, y, z };

                    let point_idx: usize = (count % (total_points as u64)).try_into()?;
                    Ok(MDDFrame {
                        frame_idx,
                        point_idx,
                        time,
                        point,
                    })
                },
            ),
        );
        Ok(frames)
    }

    fn map_to<F>(mut infile: File, mut outfile: File, mut op: F) -> Result<(), Box<dyn Error>>
    where
        F: FnMut(MDDFrame) -> Result<MDDFrame, Box<dyn Error>>,
    {
        let (total_frames, total_points) = MDDSeekableFile::read_header(&mut infile)?;
        MDDSeekableFile::write_header(&mut outfile, total_frames, total_points)?;

        let header_size: u64 = (size_of::<i32>() * 2).try_into()?;
        let f32_size: u64 = size_of::<f32>().try_into()?;
        let mut count = 0;

        for frame in Self::read_frames(infile)? {
            let frame = op(frame?)?;
            let frame_idx: u64 = count / (total_points as u64);
            let time_offset: u64 = f32_size * frame_idx;
            let time_size: u64 = f32_size * (total_frames as u64);
            let data_offset: u64 = f32_size * count * 3;

            outfile.seek(SeekFrom::Start(header_size + time_offset))?;
            outfile.write_all(&frame.time.to_be_bytes())?;

            outfile.seek(SeekFrom::Start(header_size + time_size + data_offset))?;
            outfile.write_all(&frame.point.x.to_be_bytes())?;
            outfile.write_all(&frame.point.y.to_be_bytes())?;
            outfile.write_all(&frame.point.z.to_be_bytes())?;

            count += 1;
        }

        Ok(())
    }

    fn read_animation_info(mut infile: File) -> Result<AnimationInfo, Box<dyn Error>> {
        let (total_frames, total_points) = Self::read_header(&mut infile)?;
        let mut frames = Self::read_frames(infile)?;
        let mut fps: f32 = -1.0;
        let mut prev_frame = frames.next().ok_or(EmptyAnimationError)??;
        let start_frame = prev_frame.frame_idx;
        // TODO: sometimes the time difference between frames is 0, not sure why
        // Need to rely on later frames to avoid divide by zero
        for next_frame in frames {
            let next_frame = next_frame?;
            if fps < 0.0 && (next_frame.time - prev_frame.time) > 0.0 {
                fps = 1.0 / (next_frame.time - prev_frame.time);
            }
            prev_frame = next_frame;
        }
        let end_frame = prev_frame.frame_idx;

        Ok(AnimationInfo {
            total_frames: total_frames as usize,
            total_points: total_points as usize,
            start_frame,
            end_frame,
            fps,
        })
    }
}

pub fn add(left: usize, right: usize) -> usize {
    left + right
}

#[cfg(test)]
mod tests {
    use super::*;

    #[test]
    fn it_works() -> Result<(), Box<dyn Error>> {
        let infile = File::open("test_cube.mdd")?;
        let outfile = File::create("output.mdd")?;
        println!("MAPPING test_cube.mdd to output.mdd");
        MDDSeekableFile::map_to(infile, outfile, |mut frame| {
            frame.point.x = frame.frame_idx as f32;
            frame.point.y = (frame.frame_idx * 2) as f32;
            frame.point.z = 0.0;

            println!("{:?}", frame);
            Ok(frame)
        })?;
        println!("\nReading back output.mdd");
        let infile = File::open("output.mdd")?;
        for frame in MDDSeekableFile::read_frames(infile)? {
            println!("{:?}", frame);
        }
        Ok(())
    }
    #[test]
    fn animation_info() -> Result<(), Box<dyn Error>> {
        let infile = File::open("test_cube.mdd")?;
        let animation = MDDSeekableFile::read(infile)?;
        assert_eq!(
            animation.info,
            AnimationInfo {
                total_frames: 4,
                total_points: 1,
                start_frame: 0,
                end_frame: 3,
                fps: 24.0
            }
        );
        Ok(())
    }
}
