// Creates two image files for a given [basename].mdd file
// First image contains index of every active vertex in the bigger image file, vertices that don't move have a value of -1 in the mask image.

// Openexr library is too complex to go off documentation alone and no projects on github using library. Seek out test cases
// Reference for outputting an image
// https://github.com/vfx-rs/openexr-rs/blob/25826b4f89bc768b565ba150d6f9c76876ad6bc3/src/core/output_file.rs#L275

use openexr::core::channel_list::Channel;
use openexr::core::frame_buffer::{FrameBuffer, Slice};
use openexr::core::header::Header;
use openexr::core::output_file::OutputFile;
use openexr::core::{Compression, PixelType};
use pointcache::{AnimationInfo, MDDFrame, MDDSeekableFile, Point, PointCache};
use std::io::Write;
use std::{error::Error, fs::File};

pub const MAX_IMAGE_DIM: usize = 16384;
pub const MOVEMENT_EPSILON: f32 = 0.000000001;

fn non_color_channel() -> Channel {
    Channel {
        type_: PixelType::Float.into(),
        x_sampling: 1,
        y_sampling: 1,
        // The docs say this is only used by a particular compression algorithm, but I don't think it hurts to have it here
        p_linear: false,
    }
}

// Using f32 instead of Uint to store integers because Godot does not support Uint
// Currently only using R channel, can squeeze more into 16k texture if use more channels
fn write_mask_image(
    filename: &str,
    pixels: &Vec<f32>,
    vertex_total: usize,
) -> Result<(), Box<dyn Error>> {
    let width = vertex_total.min(MAX_IMAGE_DIM);
    let height: usize = ((vertex_total as f32) / (MAX_IMAGE_DIM as f32)).ceil() as usize;
    assert!(height <= MAX_IMAGE_DIM);
    let mut header = Header::from_dimensions(width as i32, height as i32);

    header.set_compression(Compression::No);

    header.channels_mut().insert("R", &non_color_channel());

    let mut frame_buffer = FrameBuffer::new();

    frame_buffer.insert(
        "R",
        &Slice::builder(
            PixelType::Float,
            &pixels[0] as *const _ as *const u8,
            width as i64,
            height as i64,
        )
        .build()?,
    )?;

    let mut file = OutputFile::new(filename, &header, 1)?;
    file.set_frame_buffer(&frame_buffer)?;
    unsafe { file.write_pixels(height as i32)? };

    Ok(())
}

fn write_point_image(
    filename: &str,
    pixels: &Vec<Point>,
    width: usize,
    height: usize,
) -> Result<(), Box<dyn Error>> {
    let mut header = Header::from_dimensions(width as i32, height as i32);

    header.set_compression(Compression::No);

    header.channels_mut().insert("R", &non_color_channel());
    header.channels_mut().insert("G", &non_color_channel());
    header.channels_mut().insert("B", &non_color_channel());

    let mut frame_buffer = FrameBuffer::new();

    [
        ("R", &pixels[0].x),
        ("G", &pixels[0].y),
        ("B", &pixels[0].z),
    ]
    .map(|(name, spot)| {
        frame_buffer.insert(
            name,
            &Slice::builder(
                PixelType::Float,
                spot as *const _ as *const u8,
                width as i64,
                height as i64,
            )
            .x_stride(std::mem::size_of::<Point>())
            .build()?,
        )
    })
    .into_iter()
    .collect::<Result<(), openexr::Error>>()?;

    let mut file = OutputFile::new(filename, &header, 1)?;
    file.set_frame_buffer(&frame_buffer)?;
    unsafe { file.write_pixels(height as i32)? };

    Ok(())
}

// Not used because blender always starts frame frame zero so this info is useless
fn _write_animation_info(filename: &str, info: AnimationInfo) -> Result<(), Box<dyn Error>> {
    let data = format!(
        "{{ \"total_frames\": {}, \"total_points\": {}, \"start_frame\": {}, \"end_frame\": {}, \"fps\": {} }}\n",
        info.total_frames, info.total_points, info.start_frame, info.end_frame, info.fps
    );
    let mut file = File::create(filename)?;
    file.write_all(data.as_bytes())?;
    Ok(())
}

#[derive(Debug, Clone, Copy)]
enum PointState {
    Unseen,
    OneValue { point_idx: usize, point: Point },
    Varying,
    Reindexed { new_idx: usize },
}
use PointState::*;
impl PointState {
    fn mask_value(&self) -> f32 {
        match *self {
            Unseen => panic!("no point should be unseen"),
            Varying => panic!("no point should be unindexed"),
            OneValue { .. } => -1.0,
            Reindexed { new_idx, .. } => new_idx as f32,
        }
    }
    fn point_changed(a: &Point, b: &Point) -> bool {
        ((a.x - b.x).abs() + (a.y - b.y).abs() + (a.z - b.z).abs()) > MOVEMENT_EPSILON
    }
}

fn main() {
    let basename = std::env::args()
        .nth(1)
        .expect("expected first argument to be the basename to an .mdd file in the CWD");

    let mdd_filename = basename.clone() + ".mdd";

    let mdd_file = File::open(mdd_filename).expect("Mdd filename is not valid");

    let info = MDDSeekableFile::read_animation_info(mdd_file.try_clone().unwrap()).unwrap();

    let mut mask_last: Vec<PointState> = vec![Unseen; info.total_points];
    let pc = MDDSeekableFile::read_frames(mdd_file.try_clone().unwrap()).unwrap();
    for frame in pc {
        let MDDFrame {
            point, point_idx, ..
        } = frame.unwrap();
        match mask_last[point_idx] {
            Unseen => mask_last[point_idx] = OneValue { point_idx, point },
            OneValue {
                point_idx: old_point_idx,
                point: old_point,
            } => {
                assert_eq!(old_point_idx, point_idx);
                if PointState::point_changed(&point, &old_point) {
                    mask_last[point_idx] = Varying;
                }
            }
            Varying => {}
            Reindexed { .. } => unreachable!()
        }
    }

    let mut total_varying = 0;
    for point in mask_last.iter_mut() {
        if let Varying = point {
            *point = Reindexed { new_idx: total_varying };
            total_varying += 1;
        }
    }    

    // Check that the varing indices are increasing w.r.t. point_idx
    let mut mask_iter = mask_last.iter().filter_map(|x| {
        if let &Reindexed { new_idx } = x {
            Some(new_idx)
        } else {
            None
        }
    });
    if let Some(mut prev_idx) = mask_iter.next() {
        for next_idx in mask_iter {
            assert_eq!(prev_idx + 1, next_idx);
            prev_idx = next_idx;
        }
    }

    let mask: Vec<f32> = mask_last.iter().map(PointState::mask_value).collect();

    write_mask_image(&(basename.clone() + "_mask.exr"), &mask, info.total_points).unwrap();

    let mut pixels: Vec<Point> = Vec::with_capacity(total_varying * (info.total_frames));

    let pc = MDDSeekableFile::read_frames(mdd_file.try_clone().unwrap()).unwrap();
    let pc: Vec<_> = pc.collect();
    let max_point_idx = pc
        .iter()
        .map(|frame| frame.as_ref().unwrap())
        .map(|MDDFrame { point_idx, .. }| point_idx)
        .cloned()
        .max()
        .unwrap();
    assert_eq!(max_point_idx + 1, info.total_points);

    let pc = pc.into_iter();

    pixels.extend(pc.filter_map(|frame| {
        let MDDFrame {
            point_idx, point, ..
        } = frame.unwrap();
        match mask_last[point_idx] {
            Varying | Unseen => panic!(),
            OneValue { .. } => None,
            Reindexed { .. } => Some(point),
        }
    }));

    assert!(pixels.len() == total_varying * (info.total_frames));

    if total_varying == 0 {
        eprintln!("No varying frames, not outputting .exr image");
        return;
    }

    write_point_image(
        &(basename.clone() + ".exr"),
        &pixels,
        total_varying,
        info.total_frames,
    )
    .unwrap();

    println!(
        "Total {} varying points of {} total points",
        total_varying, info.total_points
    );
}
